from __future__ import annotations

from dataclasses import dataclass

LANGUAGE_CODE_SUFFIX = '_(((([A-Za-z]{2,3}(-([A-Za-z]{3}(-[A-Za-z]{3}){0,2}))?)|[A-Za-z]{4}|[A-Za-z]{5,8})(-([A-Za-z]{4}))?(-([A-Za-z]{2}|[0-9]{3}))?(-([A-Za-z0-9]{5,8}|[0-9][A-Za-z0-9]{3}))*(-([0-9A-WY-Za-wy-z](-[A-Za-z0-9]{2,8})+))*(-(x(-[A-Za-z0-9]{1,8})+))?)|(x(-[A-Za-z0-9]{1,8})+))'  # noqa: E501
LANGUAGE_CODE_SUFFIX_LEN = len(LANGUAGE_CODE_SUFFIX)


@dataclass
class Field:
    """Initialize a schema field object."""

    #: The field's name.
    name: str
    #: The field's schema.
    schema: dict
    #: The ``deprecated`` property of the field.
    deprecated_self: dict
    #: The ``deprecated`` property of the field, or an ancestor of the field.
    deprecated: dict
    #: The JSON pointer to the field in the schema, e.g. ``/properties/tender/properties/id``.
    #: Used, for example, to look up a modified field's original schema in the release schema.
    pointer: str
    #: The path to the field in data, e.g. ``('tender', 'id')``.
    path_components: tuple
    #: The definition in which the field is defined, e.g. ``'Item'``.
    definition: str
    #: Whether the field is defined under ``patternProperties``.
    pattern: bool = False
    #: Whether the field has a corresponding field in the schema's ``patternProperties`` (OCDS 1.1).
    multilingual: bool = False
    #: Whether the field is listed under ``required``.
    required: bool = False
    #: Whether the field's name is ``id`` and isn't under a ``wholeListMerge`` array.
    merge_by_id: bool = False
    #: The separator to use in string representations of paths.
    sep = '.'

    @property
    def path(self):
        """Return the path to the field in data with ``self.sep`` as separator, e.g. ``tender.id``."""
        return self.sep.join(self.path_components)

    def __repr__(self):
        return repr(self.asdict())

    def asdict(self, sep=None, exclude=()):
        """
        Return the field as a dict, with keys for all properties except ``path_components``.

        :param list sep: the separator to use in string representations of paths, overriding ``self.sep``
        :param list exclude: a list of keys to exclude from the dict
        """
        sep = sep or self.sep

        return (
            {k: v for k, v in self.__dict__.items() if k not in exclude and k != 'path_components'}
            | ({} if 'path' in exclude else {'path': sep.join(self.path_components)})
        )


def get_schema_fields(
    schema: dict,
    pointer: str = '',
    path_components: tuple = (),
    definition: str = '',
    deprecated: dict | None = None,
    *,
    whole_list_merge: bool = False,
    array: bool = False,
):
    """
    Yield a ``Field`` object for each field (whether under ``properties`` or ``patternProperties``) in a JSON schema.

    :param schema: a JSON schema
    :param pointer: the JSON pointer to the field in the schema, e.g. ``/properties/tender/properties/id``
    :param path_components: the path to the field in data, e.g. ``('tender', 'id')``
    :param definition: the definition in which the field is defined, e.g. ``'Item'``
    :param deprecated: if the field, or an ancestor of the field, sets ``deprecated``, the ``deprecated`` object
    :param whole_list_merge: whether the field, or an ancestor of the field, sets ``wholelistMerge``
    :param array: whether the field is on an entry in an array
    """
    multilingual = set()
    nonmultilingual_pattern_properties = {}

    required = schema.get('required', [])
    # ``deprecated`` and ``whole_list_merge`` are inherited.
    deprecated = deprecated or _deprecated(schema)
    whole_list_merge = whole_list_merge or schema.get('wholeListMerge', False)

    if pattern_properties := schema.get('patternProperties'):
        for pattern, subschema in pattern_properties.items():
            # The pattern might have an extra set of parentheses (OCDS 1.1). Assumes the final character is $.
            for offset in (2, 1):
                end = -LANGUAGE_CODE_SUFFIX_LEN - offset
                # The pattern must be anchored and the suffix must occur at the end.
                if (
                    pattern[end:-offset] == LANGUAGE_CODE_SUFFIX
                    and pattern[:offset] == '^('[:offset]
                    and pattern[-offset:] == ')$'[-offset:]
                ):
                    multilingual.add(pattern[offset:end])
                    break
            # Set ``multilingual`` on corresponding ``properties``. Yield remaining ``patternProperties``.
            else:
                nonmultilingual_pattern_properties[pattern] = subschema

    if items := schema.get('items'):
        # ``items`` advances the pointer and sets array context (for the next level only).
        yield from get_schema_fields(
            items,
            f'{pointer}/items',
            path_components,
            definition,
            deprecated,
            whole_list_merge=whole_list_merge,
            array=True,
        )

    for keyword in ('anyOf', 'allOf', 'oneOf'):
        if elements := schema.get(keyword):
            for i, subschema in enumerate(elements):
                # These keywords advance the pointer.
                yield from get_schema_fields(
                    subschema,
                    f'{pointer}/{keyword}/{i}',
                    path_components,
                    definition,
                    deprecated,
                    whole_list_merge=whole_list_merge,
                )

    for keyword in ('then', 'else'):
        if subschema := schema.get(keyword):
            # These keywords advance the pointer.
            yield from get_schema_fields(
                subschema,
                f'{pointer}/{keyword}',
                path_components,
                definition,
                deprecated,
                whole_list_merge=whole_list_merge,
            )

    if properties := schema.get('properties'):
        for name, subschema in properties.items():
            prop_pointer = f'{pointer}/properties/{name}'
            prop_path_components = (*path_components, name)
            prop_deprecated = _deprecated(subschema)

            yield Field(
                name=name,
                schema=subschema,
                pointer=prop_pointer,
                path_components=prop_path_components,
                definition=definition,
                deprecated_self=prop_deprecated,
                deprecated=deprecated or prop_deprecated,
                multilingual=name in multilingual,
                required=name in required,
                merge_by_id=name == 'id' and array and not whole_list_merge,
            )

            # ``properties`` advances the pointer and path.
            yield from get_schema_fields(
                subschema,
                prop_pointer,
                prop_path_components,
                definition,
                deprecated,
                whole_list_merge=whole_list_merge,
            )

    # Yield ``patternProperties`` last, to be interpreted in the context of ``properties``.
    for name, subschema in nonmultilingual_pattern_properties.items():
        prop_pointer = f'{pointer}/patternProperties/{name}'
        prop_path_components = (*path_components, name)
        prop_deprecated = _deprecated(subschema)

        yield Field(
            name=name,
            schema=subschema,
            pointer=prop_pointer,
            path_components=prop_path_components,
            definition=definition,
            deprecated_self=prop_deprecated,
            deprecated=deprecated or prop_deprecated,
            pattern=True,
            # ``patternProperties`` can't be multilingual, required, or "id".
        )

        # ``patternProperties`` advances the ppinter and path.
        yield from get_schema_fields(
            subschema,
            prop_pointer,
            prop_path_components,
            definition,
            deprecated,
            whole_list_merge=whole_list_merge,
        )

    # ``definitions`` is canonically only at the top level.
    if not pointer:
        # Yield definitions last, to be interpreted in the context of other top-level properties.
        for keyword in ('definitions', '$defs'):
            if definitions := schema.get(keyword):
                for name, subschema in definitions.items():
                    # These keywords advance the pointer and set the definition.
                    yield from get_schema_fields(subschema, f'/{keyword}/{name}', definition=name)


def _deprecated(value):
    return value.get('deprecated') or (hasattr(value, '__reference__') and value.__reference__.get('deprecated')) or {}


def add_validation_properties(schema, *, unique_items=True, coordinates=False):
    """
    Add "minItems" and "uniqueItems" if an array, add "minProperties" if an object, and add "minLength" if a string
    and if "enum", "format" and "pattern" aren't set.

    :param dict schema: a JSON schema
    :param bool unique_items: whether to add "uniqueItems" properties to array fields
    :param bool coordinates: whether the parent is a geospatial coordinates field
    """
    if isinstance(schema, list):
        for item in schema:
            add_validation_properties(item, unique_items=unique_items)
    elif isinstance(schema, dict):
        if 'type' in schema:
            if (
                'string' in schema['type']
                # "enum" is more strict than "minLength".
                and 'enum' not in schema
                # The defined formats do not match zero-length strings.
                # https://datatracker.ietf.org/doc/html/draft-fge-json-schema-validation-00#section-7.3
                and 'format' not in schema
                # The pattern is assumed to not match zero-length strings.
                and 'pattern' not in schema
            ):
                schema.setdefault('minLength', 1)

            if 'array' in schema['type']:
                # Allow non-unique items for coordinates fields (e.g. closed polygons).
                if sorted(schema.get('items', {}).get('type', [])) == ['array', 'number']:
                    coordinates = True
                if unique_items and not coordinates:
                    schema.setdefault('uniqueItems', True)
                schema.setdefault('minItems', 1)

            if 'object' in schema['type']:
                schema.setdefault('minProperties', 1)

        for value in schema.values():
            add_validation_properties(value, unique_items=unique_items, coordinates=coordinates)



import json
import os
from unittest.mock import Mock

import pytest
from libcove.lib.common import (
    _get_schema_deprecated_paths,
    _get_schema_non_required_ids,
    schema_dict_fields_generator,
)

from ocdskit.schema import get_schema_fields
from tests import load

# Remove parts of paths to reduce repetition in EXCEPTIONS.
REMOVE = {"compiledRelease", "records", "releases", "value", "versionedRelease"}

EXCEPTIONS = {
    # guatecompras/ocds_partyDetails_publicEntitiesLevelDetails_extension sets ``type`` to "object" but sets ``items``.
    ("complaints", "intervenients", "details", "entityType", "id"),
    ("complaints", "intervenients", "details", "legalEntityTypeDetail", "id"),
    ("complaints", "intervenients", "details", "level", "id"),
    ("complaints", "intervenients", "details", "type", "id"),
    ("parties", "details", "entityType", "id"),
    ("parties", "details", "legalEntityTypeDetail", "id"),
    ("parties", "details", "level", "id"),
    ("parties", "details", "type", "id"),
}


@pytest.fixture(params=["a", "b"])
def scenario(request):
    return request.param


@pytest.fixture(params=["release", "record"])
def package_type(request):
    return request.param


# For comparing values to JSON fixtures (tuples versus lists).
def dumpload(data):
    return json.loads(json.dumps(data))


def old(schema, mock):
    set(schema_dict_fields_generator(schema))
    _get_schema_non_required_ids(mock)
    _get_schema_deprecated_paths(mock)


def new(schema):
    fields = list(get_schema_fields(schema))

    # Schema fields.
    {f"/{'/'.join(field.path_components)}" for field in fields}
    # Optional IDs.
    {field.path_components for field in fields if field.merge_by_id and not field.required}
    # Deprecated fields.
    {
        field.path_components: [field.deprecated["deprecatedVersion"], field.deprecated["description"]]
        for field in fields
        if field.deprecated
    }


def test_benchmark(scenario, package_type, benchmark):
    schema = load(scenario, f"{package_type}-package-schema-dereferenced.json")
    mock = Mock()
    mock.get_pkg_schema_obj = Mock(return_value=schema)

    if os.getenv("BENCHMARK_SAVE"):
        benchmark(old, schema, mock)
    else:
        benchmark(new, schema)


# libcove: get_additional_fields_info uses get_pkg_schema_fields which calls schema_dict_fields_generator, and calls
# get_fields_present_with_examples, which calls fields_present_generator.
def test_schema_dict_fields_generator(scenario, package_type):
    schema = load(scenario, f"{package_type}-package-schema-dereferenced.json")

    expected = dumpload(sorted(set(schema_dict_fields_generator(schema))))

    actual = {
        f"/{'/'.join(field.path_components)}"
        for field in get_schema_fields(schema)
        # libcove doesn't support ``patternProperties``.
        if not field.pattern
    }

    assert expected == load(scenario, f"{package_type}-additional.json")
    assert actual == set(expected)


# libcove: get_json_data_missing_ids calls _get_schema_non_required_ids and get_json_data_generic_paths.
def test_get_schema_non_required_ids(scenario, package_type):
    schema = load(scenario, f"{package_type}-package-schema-dereferenced.json")
    mock = Mock()
    mock.get_pkg_schema_obj = Mock(return_value=schema)

    expected = dumpload(_get_schema_non_required_ids(mock))

    actual = {
        field.path_components
        for field in get_schema_fields(schema)
        if field.merge_by_id
        and not field.required
        # libcove doesn't support ``oneOf``.
        and field.path_components[:2] != ("records", "releases")
        # libcove trusts ``type``, instead of using ``properties`` and ``items``.
        and tuple(c for c in field.path_components if c not in REMOVE) not in EXCEPTIONS
    }

    assert expected == load(scenario, f"{package_type}-missing-ids.json")
    assert actual == {tuple(components) for components in expected}


# libcove: get_json_data_deprecated_fields calls _get_schema_deprecated_paths and get_json_data_generic_paths.
def test_get_schema_deprecated_paths(scenario, package_type):
    schema = load(scenario, f"{package_type}-package-schema-dereferenced.json")
    mock = Mock()
    mock.get_pkg_schema_obj = Mock(return_value=schema)

    expected = dumpload(_get_schema_deprecated_paths(mock))

    actual = {
        field.path_components: [field.deprecated_self["deprecatedVersion"], field.deprecated_self["description"]]
        for field in get_schema_fields(schema)
        if field.deprecated
        # libcove doesn't support ``oneOf``.
        and field.path_components[:2] != ("records", "releases")
        # libcove doesn't inherit deprecation.
        and field.deprecated_self
    }

    assert expected == load(scenario, f"{package_type}-deprecated.json")
    assert actual == {tuple(components): value for components, value in expected}



from ocdskit.schema import get_schema_fields
from tests import load


def test_bods_0_4():
    schema = load("bods", "schema-0-4-0.json")

    actual = {field.path_components for field in get_schema_fields(schema) if not field.definition}

    # 139
    assert actual == {
        ("annotations", "createdBy"),
        ("annotations", "createdBy", "name"),
        ("annotations", "createdBy", "uri"),
        ("annotations", "creationDate"),
        ("annotations", "description"),
        ("annotations", "motivation"),
        ("annotations", "statementPointerTarget"),
        ("annotations", "transformedContent"),
        ("annotations", "url"),
        ("annotations",),
        ("declaration",),
        ("declarationSubject",),
        ("publicationDetails", "bodsVersion"),
        ("publicationDetails", "license"),
        ("publicationDetails", "publicationDate"),
        ("publicationDetails", "publisher"),
        ("publicationDetails", "publisher", "name"),
        ("publicationDetails", "publisher", "url"),
        ("publicationDetails",),
        ("recordDetails", "addresses"),
        ("recordDetails", "addresses", "address"),
        ("recordDetails", "addresses", "country"),
        ("recordDetails", "addresses", "country", "code"),
        ("recordDetails", "addresses", "country", "name"),
        ("recordDetails", "addresses", "postCode"),
        ("recordDetails", "addresses", "type"),
        ("recordDetails", "alternateNames"),
        ("recordDetails", "birthDate"),
        ("recordDetails", "componentRecords"),
        ("recordDetails", "deathDate"),
        ("recordDetails", "dissolutionDate"),
        ("recordDetails", "entityType"),
        ("recordDetails", "entityType", "details"),
        ("recordDetails", "entityType", "subtype"),
        ("recordDetails", "entityType", "type"),
        ("recordDetails", "formedByStatute"),
        ("recordDetails", "formedByStatute", "date"),
        ("recordDetails", "formedByStatute", "name"),
        ("recordDetails", "foundingDate"),
        ("recordDetails", "identifiers"),
        ("recordDetails", "identifiers", "id"),
        ("recordDetails", "identifiers", "scheme"),
        ("recordDetails", "identifiers", "schemeName"),
        ("recordDetails", "identifiers", "uri"),
        ("recordDetails", "interestedParty"),
        ("recordDetails", "interestedParty", "description"),
        ("recordDetails", "interestedParty", "reason"),
        ("recordDetails", "interests"),
        ("recordDetails", "interests", "beneficialOwnershipOrControl"),
        ("recordDetails", "interests", "details"),
        ("recordDetails", "interests", "directOrIndirect"),
        ("recordDetails", "interests", "endDate"),
        ("recordDetails", "interests", "share"),
        ("recordDetails", "interests", "share", "exact"),
        ("recordDetails", "interests", "share", "exclusiveMaximum"),
        ("recordDetails", "interests", "share", "exclusiveMinimum"),
        ("recordDetails", "interests", "share", "maximum"),
        ("recordDetails", "interests", "share", "minimum"),
        ("recordDetails", "interests", "startDate"),
        ("recordDetails", "interests", "type"),
        ("recordDetails", "isComponent"),
        ("recordDetails", "jurisdiction"),
        ("recordDetails", "jurisdiction", "code"),
        ("recordDetails", "jurisdiction", "name"),
        ("recordDetails", "name"),
        ("recordDetails", "names"),
        ("recordDetails", "names", "familyName"),
        ("recordDetails", "names", "fullName"),
        ("recordDetails", "names", "givenName"),
        ("recordDetails", "names", "patronymicName"),
        ("recordDetails", "names", "type"),
        ("recordDetails", "nationalities"),
        ("recordDetails", "nationalities", "code"),
        ("recordDetails", "nationalities", "name"),
        ("recordDetails", "personType"),
        ("recordDetails", "placeOfBirth"),
        ("recordDetails", "placeOfBirth", "address"),
        ("recordDetails", "placeOfBirth", "country"),
        ("recordDetails", "placeOfBirth", "country", "code"),
        ("recordDetails", "placeOfBirth", "country", "name"),
        ("recordDetails", "placeOfBirth", "postCode"),
        ("recordDetails", "placeOfBirth", "type"),
        ("recordDetails", "politicalExposure"),
        ("recordDetails", "politicalExposure", "details"),
        ("recordDetails", "politicalExposure", "details", "endDate"),
        ("recordDetails", "politicalExposure", "details", "jurisdiction"),
        ("recordDetails", "politicalExposure", "details", "jurisdiction", "code"),
        ("recordDetails", "politicalExposure", "details", "jurisdiction", "name"),
        ("recordDetails", "politicalExposure", "details", "missingInfoReason"),
        ("recordDetails", "politicalExposure", "details", "reason"),
        ("recordDetails", "politicalExposure", "details", "source"),
        ("recordDetails", "politicalExposure", "details", "source", "assertedBy"),
        ("recordDetails", "politicalExposure", "details", "source", "assertedBy", "name"),
        ("recordDetails", "politicalExposure", "details", "source", "assertedBy", "uri"),
        ("recordDetails", "politicalExposure", "details", "source", "description"),
        ("recordDetails", "politicalExposure", "details", "source", "retrievedAt"),
        ("recordDetails", "politicalExposure", "details", "source", "type"),
        ("recordDetails", "politicalExposure", "details", "source", "url"),
        ("recordDetails", "politicalExposure", "details", "startDate"),
        ("recordDetails", "politicalExposure", "status"),
        ("recordDetails", "publicListing"),
        ("recordDetails", "publicListing", "companyFilingsURLs"),
        ("recordDetails", "publicListing", "hasPublicListing"),
        ("recordDetails", "publicListing", "securitiesListings"),
        ("recordDetails", "publicListing", "securitiesListings", "marketIdentifierCode"),
        ("recordDetails", "publicListing", "securitiesListings", "operatingMarketIdentifierCode"),
        ("recordDetails", "publicListing", "securitiesListings", "security"),
        ("recordDetails", "publicListing", "securitiesListings", "security", "id"),
        ("recordDetails", "publicListing", "securitiesListings", "security", "idScheme"),
        ("recordDetails", "publicListing", "securitiesListings", "security", "ticker"),
        ("recordDetails", "publicListing", "securitiesListings", "stockExchangeJurisdiction"),
        ("recordDetails", "publicListing", "securitiesListings", "stockExchangeName"),
        ("recordDetails", "subject"),
        ("recordDetails", "subject", "description"),
        ("recordDetails", "subject", "reason"),
        ("recordDetails", "taxResidencies"),
        ("recordDetails", "taxResidencies", "code"),
        ("recordDetails", "taxResidencies", "name"),
        ("recordDetails", "unspecifiedEntityDetails"),
        ("recordDetails", "unspecifiedEntityDetails", "description"),
        ("recordDetails", "unspecifiedEntityDetails", "reason"),
        ("recordDetails", "unspecifiedPersonDetails"),
        ("recordDetails", "unspecifiedPersonDetails", "description"),
        ("recordDetails", "unspecifiedPersonDetails", "reason"),
        ("recordDetails", "uri"),
        ("recordDetails",),
        ("recordId",),
        ("recordStatus",),
        ("recordType",),
        ("source", "assertedBy"),
        ("source", "assertedBy", "name"),
        ("source", "assertedBy", "uri"),
        ("source", "description"),
        ("source", "retrievedAt"),
        ("source", "type"),
        ("source", "url"),
        ("source",),
        ("statementDate",),
        ("statementId",),
    }


def test_bods_0_3():
    schema = load("bods", "schema-0-3-0.json")

    actual = {field.path_components for field in get_schema_fields(schema)}

    # 138
    assert actual == {
        ("addresses", "address"),
        ("addresses", "country"),
        ("addresses", "postCode"),
        ("addresses", "type"),
        ("addresses",),
        ("alternateNames",),
        ("annotations", "createdBy"),
        ("annotations", "createdBy", "name"),
        ("annotations", "createdBy", "uri"),
        ("annotations", "creationDate"),
        ("annotations", "description"),
        ("annotations", "motivation"),
        ("annotations", "statementPointerTarget"),
        ("annotations", "transformedContent"),
        ("annotations", "url"),
        ("annotations",),
        ("birthDate",),
        ("componentStatementIDs",),
        ("deathDate",),
        ("dissolutionDate",),
        ("entitySubtype", "generalCategory"),
        ("entitySubtype", "localTerm"),
        ("entitySubtype",),
        ("entityType",),
        ("formedByStatute", "date"),
        ("formedByStatute", "name"),
        ("formedByStatute",),
        ("foundingDate",),
        ("identifiers", "id"),
        ("identifiers", "scheme"),
        ("identifiers", "schemeName"),
        ("identifiers", "uri"),
        ("identifiers",),
        ("interestedParty", "describedByEntityStatement"),
        ("interestedParty", "describedByPersonStatement"),
        ("interestedParty", "unspecified"),
        ("interestedParty", "unspecified", "description"),
        ("interestedParty", "unspecified", "reason"),
        ("interestedParty",),
        ("interests", "beneficialOwnershipOrControl"),
        ("interests", "details"),
        ("interests", "directOrIndirect"),
        ("interests", "endDate"),
        ("interests", "share"),
        ("interests", "share", "exact"),
        ("interests", "share", "exclusiveMaximum"),
        ("interests", "share", "exclusiveMinimum"),
        ("interests", "share", "maximum"),
        ("interests", "share", "minimum"),
        ("interests", "startDate"),
        ("interests", "type"),
        ("interests",),
        ("isComponent",),
        ("jurisdiction", "code"),
        ("jurisdiction", "name"),
        ("jurisdiction",),
        ("name",),
        ("names", "familyName"),
        ("names", "fullName"),
        ("names", "givenName"),
        ("names", "patronymicName"),
        ("names", "type"),
        ("names",),
        ("nationalities", "code"),
        ("nationalities", "name"),
        ("nationalities",),
        ("personType",),
        ("placeOfBirth", "address"),
        ("placeOfBirth", "country"),
        ("placeOfBirth", "postCode"),
        ("placeOfBirth", "type"),
        ("placeOfBirth",),
        ("placeOfResidence", "address"),
        ("placeOfResidence", "country"),
        ("placeOfResidence", "postCode"),
        ("placeOfResidence", "type"),
        ("placeOfResidence",),
        ("politicalExposure", "details"),
        ("politicalExposure", "details", "endDate"),
        ("politicalExposure", "details", "jurisdiction"),
        ("politicalExposure", "details", "jurisdiction", "code"),
        ("politicalExposure", "details", "jurisdiction", "name"),
        ("politicalExposure", "details", "missingInfoReason"),
        ("politicalExposure", "details", "reason"),
        ("politicalExposure", "details", "source"),
        ("politicalExposure", "details", "source", "assertedBy"),
        ("politicalExposure", "details", "source", "assertedBy", "name"),
        ("politicalExposure", "details", "source", "assertedBy", "uri"),
        ("politicalExposure", "details", "source", "description"),
        ("politicalExposure", "details", "source", "retrievedAt"),
        ("politicalExposure", "details", "source", "type"),
        ("politicalExposure", "details", "source", "url"),
        ("politicalExposure", "details", "startDate"),
        ("politicalExposure", "status"),
        ("politicalExposure",),
        ("publicationDetails", "bodsVersion"),
        ("publicationDetails", "license"),
        ("publicationDetails", "publicationDate"),
        ("publicationDetails", "publisher"),
        ("publicationDetails", "publisher", "name"),
        ("publicationDetails", "publisher", "url"),
        ("publicationDetails",),
        ("publicListing", "companyFilingsURLs"),
        ("publicListing", "hasPublicListing"),
        ("publicListing", "securitiesListings"),
        ("publicListing", "securitiesListings", "marketIdentifierCode"),
        ("publicListing", "securitiesListings", "operatingMarketIdentifierCode"),
        ("publicListing", "securitiesListings", "security"),
        ("publicListing", "securitiesListings", "security", "id"),
        ("publicListing", "securitiesListings", "security", "idScheme"),
        ("publicListing", "securitiesListings", "security", "ticker"),
        ("publicListing", "securitiesListings", "stockExchangeJurisdiction"),
        ("publicListing", "securitiesListings", "stockExchangeName"),
        ("publicListing",),
        ("replacesStatements",),
        ("source", "assertedBy"),
        ("source", "assertedBy", "name"),
        ("source", "assertedBy", "uri"),
        ("source", "description"),
        ("source", "retrievedAt"),
        ("source", "type"),
        ("source", "url"),
        ("source",),
        ("statementDate",),
        ("statementID",),
        ("statementType",),
        ("subject", "describedByEntityStatement"),
        ("subject",),
        ("taxResidencies", "code"),
        ("taxResidencies", "name"),
        ("taxResidencies",),
        ("unspecifiedEntityDetails", "description"),
        ("unspecifiedEntityDetails", "reason"),
        ("unspecifiedEntityDetails",),
        ("unspecifiedPersonDetails", "description"),
        ("unspecifiedPersonDetails", "reason"),
        ("unspecifiedPersonDetails",),
        ("uri",),
    }
